@@ -16,6 +16,8 @@ Author: Thorsten Wagner (thorsten.wagner@mpi-dortmund.mpg.de)
 # ---------------------------------------------------------------------------
 import csv
 import os
+import typing
+
 import numpy as np
 from typing import List, Union
 import cryolo.utils as utils
@@ -133,7 +135,7 @@ def is_star_filament_file(path):
         return False
     return True
 
-def read_star_filament_file(path, box_width, min_distance=0):
+def read_star_filament_file_(path, box_width, min_distance=0):
     sfile = star.StarFile(path)
 
     xcoords = sfile['']["_rlnCoordinateX"]
@@ -160,7 +162,7 @@ def read_star_filament_file(path, box_width, min_distance=0):
             filaments.append(f)
     return filaments
 
-def write_star_filemant_file(path, filaments):
+def write_star_filemant_file_(path, filaments):
     num_boxes = 2*len(filaments)
 
     coords = np.zeros(shape=(num_boxes,2))
@@ -209,6 +211,94 @@ def read_cbox_include_list(path):
     starfile = star.StarFile(path)
     return starfile['cryolo_include']["_slice_index"].tolist()
 
+def read_star_filament_file(path, box_width, min_distance=0) -> typing.List[Filament]:
+    sfile = star.StarFile(path)
+    filaments = {}
+    for i in range(len(sfile['']['_rlnCoordinateX'])):
+        box = BoundBox(
+            x=sfile['']['_rlnCoordinateX'][i]-box_width/2,
+            y=sfile['']['_rlnCoordinateY'][i]-box_width/2,
+            w=box_width,
+            h=box_width,
+            c=sfile['']['_rlnAutopickFigureOfMerit'][i]
+        )
+
+
+        fid = int(sfile['']['_rlnHelicalTubeID'][i])
+        if fid in filaments:
+            filaments[fid].append(box)
+        else:
+            filaments[fid] = [box]
+    filaments_list = [Filament(filaments[key]) for key in filaments]
+    if min_distance > 0:
+        filaments_list = utils.resample_filaments(filaments_list, min_distance)
+
+    return filaments_list
+
+
+def write_star_filemant_file(path, filaments):
+    columns = ['_rlnCoordinateX', '_rlnCoordinateY', '_rlnAutopickFigureOfMerit', '_rlnHelicalTubeID', '_rlnAnglePsiPrior']
+    data_dict = {}
+    for clm in columns:
+        data_dict[clm] = []
+
+    for fil_id, fil in enumerate(filaments):
+        utils.set_filament_psi_priors(fil)
+        for box_index, box in enumerate(fil.boxes):
+            data_dict['_rlnCoordinateX'].append(box.x + box.w / 2)
+            data_dict['_rlnCoordinateY'].append(box.y + box.h / 2)
+            data_dict['_rlnAutopickFigureOfMerit'].append(box.c)
+            data_dict['_rlnHelicalTubeID'].append(fil_id + 1)
+            data_dict['_rlnAnglePsiPrior'].append(fil.meta['psi'][box_index])
+
+    df = pd.DataFrame(data_dict)
+    sfile = star.StarFile(path)
+    sfile.update('', df, True)
+    sfile.write_star_file(overwrite=True)
+
+
+def write_cryosparc_star_file(path: str, coordinates: Union[List[BoundBox], List[Filament]], micrograph_name: str):
+    is_filament_data = isinstance(coordinates[0], Filament)
+    columns = ['_rlnMicrographName', '_rlnCoordinateX', '_rlnCoordinateY', '_rlnAutopickFigureOfMerit']
+    if is_filament_data:
+        columns.append('_rlnHelicalTubeID')
+        columns.append('_rlnAnglePsiPrior')
+    data_dict = {}
+    fid_offset = 0
+    if os.path.exists(path):
+        starfile = star.StarFile(path)
+        if '_rlnHelicalTubeID' in starfile[''] and len(starfile['']['_rlnHelicalTubeID'])>0:
+            fid_offset=np.max(starfile['']['_rlnHelicalTubeID'])
+
+    for clm in columns:
+        data_dict[clm] = []
+    if is_filament_data:
+        for fil_id, fil in enumerate(coordinates):
+            utils.set_filament_psi_priors(fil)
+            m_name = micrograph_name[fil_id]
+            for box_index, box in enumerate(fil.boxes):
+                data_dict['_rlnMicrographName'].append(m_name)
+                data_dict['_rlnCoordinateX'].append(box.x + box.w / 2)
+                data_dict['_rlnCoordinateY'].append(box.y + box.h / 2)
+                data_dict['_rlnAutopickFigureOfMerit'].append(box.c)
+                data_dict['_rlnHelicalTubeID'].append(fid_offset+fil_id+1)
+                data_dict['_rlnAnglePsiPrior'].append(fil.meta['psi'][box_index])
+    else:
+        for box_index, box in enumerate(coordinates):
+            data_dict['_rlnMicrographName'].append(micrograph_name[box_index])
+            data_dict['_rlnCoordinateX'].append(box.x + box.w / 2)
+            data_dict['_rlnCoordinateY'].append(box.y + box.h / 2)
+            data_dict['_rlnAutopickFigureOfMerit'].append(box.c)
+
+    if os.path.exists(path):
+        df = starfile['']
+        df = df.append(pd.DataFrame(data_dict))
+    else:
+        df = pd.DataFrame(data_dict)
+
+    sfile = star.StarFile(path)
+    sfile.update('', df, True)
+    sfile.write_star_file(overwrite=True)
 
 
 def read_cbox_boxfile(path, min_dist_boxes_filament=0) -> Union[List[BoundBox], List[Filament]]:
@@ -259,11 +349,10 @@ def read_cbox_boxfile(path, min_dist_boxes_filament=0) -> Union[List[BoundBox],
             else:
                 return boxes
     except Exception :
-
         print("Reading old CBOX format file")
-
         return read_cbox_boxfile_old(path)
 
+
 def read_cbox_boxfile_old(path):
     """
     Read a box file in EMAN1 box format.
@@ -486,6 +575,13 @@ def write_eman1_filament_start_end(filaments, path):
                 ]
             )
 
+def is_star_filament_file(path: str) -> bool:
+    if not path.endswith(".star"):
+        return False
+
+    sfile = star.StarFile(path)
+
+    return '_rlnHelicalTubeID' in sfile['']
 
 def is_eman1_filament_start_end(path):
     if not path.endswith(".box"):
@@ -644,7 +740,6 @@ def write_eman1_helicon(filaments, path, image_filename):
                         ]
                     )
 
-                    # helix: (3597.3, 2470.9),(3110.9, 3091.7000000000003),38
                     for box in fil.boxes[1:-1]:
                         boxwriter.writerow(
                             [
@@ -653,4 +748,4 @@ def write_eman1_helicon(filaments, path, image_filename):
                                 + " "
                                 + str(box.y + boxsize / 2)
                             ]
-                        )
+                        )
\ No newline at end of file
