@@ -35,13 +35,45 @@ from cryolo.MultiGPUModelCheckpoint import MultiGPUModelCheckpoint
 from cryolo.ExtendedModelCheckpoint import ExtendedModelCheckpoint
 from cryolo.warmup_callback import WarmupCallback
 from cryolo.SGDRSchedular import SGDRScheduler
-from cryolo.backend import FullYoloFeature, CrYoloFeature, PhosaurusNetFeature
+from cryolo.backend import FullYoloFeature, CrYoloFeature, PhosaurusNetFeature, BaseFeatureExtractor
 from cryolo.patchwisebatchgenerator2 import PatchwiseBatchGenerator
 from cryolo.utils import BoundBox
-
+from typing import Sequence
 warmup_bs = keras.backend.variable(1.0)
 
 
+class FeatureExtractorFactory():
+
+    @staticmethod
+    def create_feature_extractor(architecture : str , input_size : Sequence, input_depth : int, backend_weights : str) -> BaseFeatureExtractor:
+        try:
+            if architecture == "YOLO":
+                return FullYoloFeature(
+                    input_size, input_depth, backend_weights
+                )
+            elif architecture == "crYOLO":
+                return CrYoloFeature(
+                    input_size, input_depth, backend_weights
+                )
+            elif architecture == "PhosaurusNet":
+                return PhosaurusNetFeature(
+                    input_size, input_depth, backend_weights
+                )
+            else:
+                raise Exception(
+                    "Architecture not supported! Only support YOLO and crYOLO at the moment!"
+                )
+        except tf.errors.ResourceExhaustedError as e:
+            print(e)
+            import sys
+
+            sys.exit(
+                "Seems that crYOLO was not able to allocate memory. "
+                "Is your GPU currenlty used? Check it with nivdia-smi"
+            )
+        return None
+
+
 class YOLO:
     def __init__(
         self,
@@ -53,10 +85,12 @@ class YOLO:
         uniitest=False,
         experimental_loss=False,
         pretrained_weights=None,
+        pretrained_weights_init=False, #only use for pre initialization during training
         fine_tune=False,
         num_fine_tune_layers=2,
         input_depth=1,
         labels=["particle"],
+        filament_model=False
     ):
         self.input_size = input_size
         self.input_depth = input_depth
@@ -72,6 +106,8 @@ class YOLO:
         self.features = None
         self.fine_tune = fine_tune
         self.num_fine_tune_layers = num_fine_tune_layers
+
+
         ##########################
         # Make the model
         ##########################
@@ -86,32 +122,13 @@ class YOLO:
                 shape=(self.input_size[1], self.input_size[0], self.input_depth),
                 dtype="float32",
             )
-            self.true_boxes = Input(shape=(1, 1, 1, max_box_per_image, 4))
-            try:
-                if architecture == "YOLO":
-                    self.feature_extractor = FullYoloFeature(
-                        self.input_size, self.input_depth, backend_weights
-                    )
-                elif architecture == "crYOLO":
-                    self.feature_extractor = CrYoloFeature(
-                        self.input_size, self.input_depth, backend_weights
-                    )
-                elif architecture == "PhosaurusNet":
-                    self.feature_extractor = PhosaurusNetFeature(
-                        self.input_size, self.input_depth, backend_weights
-                    )
-                else:
-                    raise Exception(
-                        "Architecture not supported! Only support YOLO and crYOLO at the moment!"
-                    )
-            except tf.errors.ResourceExhaustedError as e:
-                print(e)
-                import sys
+            self.is_filament_model = filament_model
+            self.true_boxes = Input(shape=(1, 1, 1, max_box_per_image, 4 + self.is_filament_model))
 
-                sys.exit(
-                    "Seems that crYOLO was not able to allocate memory. "
-                    "Is your GPU currenlty used? Check it with nivdia-smi"
-                )
+            self.feature_extractor = FeatureExtractorFactory.create_feature_extractor(architecture=architecture,
+                                                                                      input_size=self.input_size,
+                                                                                      input_depth=self.input_depth,
+                                                                                      backend_weights=backend_weights)
 
             self.grid_rows, self.grid_columns = self.feature_extractor.get_output_shape()
 
@@ -119,16 +136,19 @@ class YOLO:
 
             # make the object detection layer
             self.features = Dropout(0.2, name="Dropout1")(features)
+            postfix = ""
+            if self.is_filament_model:
+                postfix = "_filament"
             output = Conv2D(
-                self.nb_box * (4 + 1 + self.nb_class),
+                self.nb_box * (4 + 1 + self.nb_class + self.is_filament_model),
                 (1, 1),
                 strides=(1, 1),
                 padding="same",
-                name="conv_23",
+                name="conv_23" + postfix,
                 kernel_initializer="lecun_normal",
             )(features)
             output = Reshape(
-                (self.grid_rows, self.grid_columns, self.nb_box, 4 + 1 + self.nb_class)
+                (self.grid_rows, self.grid_columns, self.nb_box, 4 + 1 + self.nb_class + self.is_filament_model)
             )(output)
             output = Lambda(lambda args: args[0])([output, self.true_boxes])
             self.model = Model([input_image, self.true_boxes], output)
@@ -152,6 +172,7 @@ class YOLO:
                     weight_path=pretrained_weights,
                     do_fine_tune=self.fine_tune,
                     num_layers=self.num_fine_tune_layers,
+                    is_init=pretrained_weights_init
                 )
 
     def custom_loss(self, y_true, y_pred):
@@ -204,11 +225,17 @@ class YOLO:
             self.anchors, [1, 1, 1, self.nb_box, 2]
         )
 
+
+
         # adjust confidence
         pred_box_conf = tf.sigmoid(y_pred[..., 4])
 
-        # adjust class probabilities
-        pred_box_class = y_pred[..., 5:]
+        if self.is_filament_model:
+            pred_box_angle = tf.sigmoid(y_pred[..., -1]) # from 0 to 1 -> 0 to PI
+            pred_box_class = y_pred[..., 5:-1]
+        else:
+            # adjust class probabilities
+            pred_box_class = y_pred[..., 5:]
 
         #
         # Adjust ground truth
@@ -245,9 +272,15 @@ class YOLO:
         true_box_conf = iou_scores * y_true[..., 4]  # Page 2 in YOLO Paper
 
         # adjust class probabilities
-        true_box_class = tf.argmax(
-            y_true[..., 5:], -1
-        )  # is always zero if there is only one class
+        if self.is_filament_model:
+            true_box_class = tf.argmax(
+                y_true[..., 5:-1], -1
+            )
+            true_box_angle = y_true[..., -1]
+        else:
+            true_box_class = tf.argmax(
+                y_true[..., 5:], -1
+            )  # is always zero if there is only one class
 
         #
         # Determine the masks
@@ -351,6 +384,31 @@ class YOLO:
             / (nb_coord_box + 1e-6)
             / 2.0
         )
+
+        if self.is_filament_model:
+            # I should do it similar to confidence, but not sure why the
+            # division by 2 happens
+            #loss_angle = true_box_angle - pred_box_angle
+            '''
+            diff_angles_a = tf.abs(true_box_angle - pred_box_angle)
+            # Reason for +-0.5: The 0.5 is actuall +- PI, but the true angles are in the range from
+            # 0 to 1 instead of 0 to 2PI because they are normalized. 0.5 is also actually +- PI
+            diff_angles_b = tf.abs(true_box_angle - pred_box_angle + 1)
+            diff_angles_c = tf.abs(true_box_angle - pred_box_angle - 1)
+
+            diff_angles = tf.minimum(tf.math.minimum(diff_angles_a,diff_angles_b),diff_angles_c)
+            '''
+            dot = tf.square(
+                tf.cos(true_box_angle*np.pi)*tf.cos(pred_box_angle*np.pi)+
+                tf.sin(true_box_angle*np.pi)*tf.sin(pred_box_angle*np.pi)
+            )
+            loss_angle = (
+                tf.reduce_sum((1-dot) * coord_mask[:,:,:,0,:]) # conf_mask
+                / (nb_coord_box + 1e-6) # nb_conf_box
+                / 2
+            )
+
+
         if self.experimental_loss:
             loss_conf_obj = (
                 tf.reduce_sum(tf.square(true_box_conf - pred_box_conf) * conf_mask_obj)
@@ -378,7 +436,10 @@ class YOLO:
         if self.experimental_loss:
             loss = loss_xy + loss_conf_obj + loss_conf_noobj + loss_class + loss_wh
         else:
-            loss = loss_xy + loss_conf + loss_class + loss_wh
+            if self.is_filament_model:
+                loss = loss_xy + loss_conf + loss_class + loss_wh + loss_angle
+            else:
+                loss = loss_xy + loss_conf + loss_class + loss_wh
 
         if self.debug:
             nb_true_box = tf.reduce_sum(y_true[..., 4])
@@ -426,7 +487,11 @@ class YOLO:
                         "\n",
                         "\tLoss Conf \t",
                         loss_conf,
-                        "\n" "\tTotal Loss \t",
+                        "\n",
+                        "\tLoss Angle \t",
+                         loss_angle,
+                        "\n",
+                        "\tTotal Loss \t",
                         loss,
                         "\n",
                         "\tCurrent Recall \t",
@@ -467,9 +532,20 @@ class YOLO:
 
         return loss
 
-    def load_weights(self, weight_path, do_fine_tune=False, num_layers=2):
+    def load_weights(self, weight_path: str, do_fine_tune: bool=False, num_layers: int=2, is_init: bool=False):
+        '''
+
+        :param weight_path: Path to weights
+        :param do_fine_tune: Boolean flag if only fine tuning is done
+        :param num_layers: Number of layers to fine tune
+        :param is_init: Boolean flag. True if the weights are only loaded for pre initialization during training.
+        '''
+
         try:
-            self.model.load_weights(weight_path)
+            if is_init:
+                self.model.load_weights(weight_path, by_name=True)
+            else:
+                self.model.load_weights(weight_path)
 
             ###############################
             # If wanted, only train the last two layers of the network.
@@ -541,6 +617,7 @@ class YOLO:
             for layer_index, layer in enumerate(self.model.get_layer(index=1).layers):
                 if layer.name not in free_layers:
                     layer.trainable = False
+
                 else:
                     print("Make trainable:", layer.name)
 
@@ -555,7 +632,7 @@ class YOLO:
         nms_threshold=0.3,
         num_patches=1,
         normalize_margin=0,
-        normalization="STANDARD"
+        normalization="STANDARD",
     ):
 
         self.nms_threshold = nms_threshold
@@ -584,7 +661,7 @@ class YOLO:
         input_image = resized_normalized_array
         input_image = np.expand_dims(input_image, 3)
         dummy_array = np.zeros(
-            (input_image.shape[0], 1, 1, 1, self.max_box_per_image, 4)
+            (input_image.shape[0], 1, 1, 1, self.max_box_per_image, 4+self.is_filament_model)
         )
         netout = self.model.predict([input_image, dummy_array])
         boxes_nms, boxes_unfiltered = self.decode_netout_fast(
@@ -616,12 +693,19 @@ class YOLO:
             netout = netout_batch[netout_index, :, :, :, :]
             grid_h, grid_w, nb_box = netout.shape[:3]
             # decode the output by the network
-            netout[..., 4] = sigm(netout[..., 4])
-            netout[..., 5:] = netout[..., 4][..., np.newaxis] * self.softmax(
-                netout[..., 5:]
-            )
+            netout[..., 4] = sigm(netout[..., 4]) # Confidence
+            if self.is_filament_model:
+                netout[..., 5:-1] = netout[..., 4][..., np.newaxis] * self.softmax(
+                    netout[..., 5:-1]
+                )
 
-            netout[..., 5:] *= netout[..., 5:] > min(0.1, obj_threshold)
+                netout[..., 5:-1] *= netout[..., 5:-1] > min(0.1, obj_threshold)
+            else:
+                netout[..., 5:] = netout[..., 4][..., np.newaxis] * self.softmax(
+                    netout[..., 5:]
+                )
+
+                netout[..., 5:] *= netout[..., 5:] > min(0.1, obj_threshold)
 
             # Transform data
             range_grid_h = np.arange(grid_h)
@@ -655,7 +739,39 @@ class YOLO:
                 self.anchors[1] * np.exp(netout[:, :, 0, 3]) / grid_h
             )  # unit: image height
             netout[:, :, 0, 3] = netout[:, :, 0, 3] * tile_shape[0] / image_dim[0]
+
+            if self.is_filament_model:
+                netout[..., 6] = sigm(netout[..., 6])*np.pi
+
             # Create boxes data
+            for row in range(grid_h):
+                for col in range(grid_w):
+                    for b in range(nb_box):
+                        if self.is_filament_model:
+                            if np.sum(netout[row, col, b, 5:-1]) > 0:
+                                bb = BoundBox(
+                                    x=netout[row, col, b, 0],
+                                    y=netout[row, col, b, 1],
+                                    w=netout[row, col, b, 2],
+                                    h=netout[row, col, b, 3],
+                                    c=netout[row, col, b, 4],
+                                    classes=netout[row, col, b, 5:-1],
+                                )
+                                bb.meta={"angle":netout[row, col, b, 6]}
+                                #print("Angle:", netout[row, col, b, 6])
+                                boxes.append(bb)
+                        else:
+                            if np.sum(netout[row, col, b, 5:]) > 0:
+                                bb = BoundBox(
+                                        x=netout[row, col, b, 0],
+                                        y=netout[row, col, b, 1],
+                                        w=netout[row, col, b, 2],
+                                        h=netout[row, col, b, 3],
+                                        c=netout[row, col, b, 4],
+                                        classes=netout[row, col, b, 5:],
+                                    )
+                                boxes.append(bb)
+            '''
             boxes.extend(
                 [
                     BoundBox(
@@ -672,20 +788,21 @@ class YOLO:
                     if np.sum(netout[row, col, b, 5:]) > 0
                 ]
             )
-
+            '''
             if (netout_index + 1) % nr_patches == 0:
                 boxes_per_image.append(boxes)
 
                 boxes = []
+
         boxes_nms = [
-            non_maxima_suppress_fast(
+            utils.non_maxima_suppress_fast(
                 boxes, self.nb_class, self.nms_threshold, min(0.1, obj_threshold)
             )
             for boxes in boxes_per_image
         ]
 
         boxes_thresholded = [
-            [copy.copy(box) for box in boxes if box.c > obj_threshold]
+            [copy.deepcopy(box) for box in boxes if box.c > obj_threshold]
             for boxes in boxes_nms
         ]
 
@@ -771,14 +888,14 @@ class YOLO:
                 boxes = []
 
         boxes_nms = [
-            non_maxima_suppress_fast(
+            utils.non_maxima_suppress_fast(
                 boxes, self.nb_class, self.nms_threshold, self.obj_threshold
             )
             for boxes in boxes_per_image
         ]
 
         boxes_less_filtered = [
-            non_maxima_suppress_fast(
+            utils.non_maxima_suppress_fast(
                 boxes, self.nb_class, self.nms_threshold, min(0.01, obj_threshold)
             )
             for boxes in boxes_per_image
@@ -824,7 +941,8 @@ class YOLO:
         num_cpus=-1,
         do_augmentation=True,
         do_multiprocessing=True,
-        normalization="STANDARD"
+        normalization="STANDARD",
+        filament_model=False
     ):
 
         self.batch_size = batch_size
@@ -913,6 +1031,7 @@ class YOLO:
             overlap=overlap_patches,
             name="train",
             train_times=train_times,
+            is_filament_training=filament_model
         )
 
         valid_batch = PatchwiseBatchGenerator(
@@ -925,6 +1044,7 @@ class YOLO:
             overlap=overlap_patches,
             name="valid",
             train_times=1,
+            is_filament_training=filament_model
         )
 
         ############################################
@@ -965,6 +1085,7 @@ class YOLO:
         checkpoint.set_anchors(self.anchors)
         checkpoint.set_fine_tune(self.fine_tune, self.num_fine_tune_layers)
         checkpoint.set_version(cryolo.__version__)
+        checkpoint.set_filamet_model(self.is_filament_model)
 
         reduceLROnPlateau = keras.callbacks.ReduceLROnPlateau(
             monitor="val_loss",
@@ -1058,65 +1179,5 @@ class YOLO:
                 use_multiprocessing=self.do_multiprocessing,
             )
         # Rename to the user specified filename
-        os.rename(saved_weights_name, saved_weights_filename + saved_weights_extension)
-
-
-def non_maxima_suppress_fast(boxes, number_classes, nms_threshold, obj_threshold):
-    for c in range(number_classes):
-        sorted_indices = list(reversed(np.argsort([box.classes[c] for box in boxes])))
-        boxes_data = np.empty(shape=(len(boxes), 7))
-        for i, sorted_index in enumerate(sorted_indices):
-            boxes_data[i, 0] = boxes[sorted_index].x
-            boxes_data[i, 1] = boxes[sorted_index].y
-            boxes_data[i, 2] = boxes[sorted_index].w
-            boxes_data[i, 3] = boxes[sorted_index].h
-            boxes_data[i, 4] = boxes[sorted_index].classes[c]
-            boxes_data[i, 5] = sorted_index
-            boxes_data[i, 6] = i
-
-        for i in range(len(boxes_data)):
-            box_i = boxes_data[i, :]
-            if box_i[4] == 0:
-                continue
-            else:
-                other_box = boxes_data[(i + 1) :]
-                if other_box.shape[0] > 0:
-                    boxes_i_rep = np.array([box_i] * len(other_box))
-                    ious = bbox_iou_vec(boxes_i_rep, other_box)
-                    boxes_data[other_box[ious >= nms_threshold, 6].astype(int), 4] = 0
-
-        for row in boxes_data[boxes_data[:, 4] == 0]:
-            boxes[row[5].astype(int)].classes[c] = 0
-
-    # remove the boxes which are less likely than a obj_threshold
-    boxes = [copy.copy(box) for box in boxes if box.get_score() > obj_threshold]
-
-    return boxes
-
-
-def bbox_iou_vec(boxesA, boxesB):
-    x1_min = boxesA[:, 0] - boxesA[:, 2] / 2
-    x1_max = boxesA[:, 0] + boxesA[:, 2] / 2
-    y1_min = boxesA[:, 1] - boxesA[:, 3] / 2
-    y1_max = boxesA[:, 1] + boxesA[:, 3] / 2
-
-    x2_min = boxesB[:, 0] - boxesB[:, 2] / 2
-    x2_max = boxesB[:, 0] + boxesB[:, 2] / 2
-    y2_min = boxesB[:, 1] - boxesB[:, 3] / 2
-    y2_max = boxesB[:, 1] + boxesB[:, 3] / 2
-    intersect_w = interval_overlap_vec(x1_min, x1_max, x2_min, x2_max)
-    intersect_h = interval_overlap_vec(y1_min, y1_max, y2_min, y2_max)
-    intersect = intersect_w * intersect_h
-    union = boxesA[:, 2] * boxesA[:, 3] + boxesB[:, 2] * boxesB[:, 3] - intersect
-    return intersect / union
-
-
-def interval_overlap_vec(x1_min, x1_max, x2_min, x2_max):
-    intersect = np.zeros(shape=(len(x1_min)))
-    cond_a = x2_min < x1_min
-    cond_b = cond_a & (x2_max >= x1_min)
-    intersect[cond_b] = np.minimum(x1_max[cond_b], x2_max[cond_b]) - x1_min[cond_b]
-    cond_c = ~cond_a & (x1_max >= x2_min)
-    intersect[cond_c] = np.minimum(x1_max[cond_c], x2_max[cond_c]) - x2_min[cond_c]
-
-    return intersect
+        if os.path.exists(saved_weights_name):
+            os.rename(saved_weights_name, saved_weights_filename + saved_weights_extension)
