@@ -20,12 +20,16 @@ import multiprocessing as mp
 import random
 import time
 from PIL import Image
+import cryolo.utils as utils
 
 import numpy as np
 from keras.utils import Sequence
 
 import cryolo.imagereader as imagereader
-from cryolo.augmentation import Augmentation
+from cryolo.augmentation.augmentation import Augmentation
+from cryolo.augmentation.FlipAugmentation import FlipAugmentation
+from cryolo.augmentation.CustomRotationAugmentation import CustomRotationAugmentation
+from cryolo.augmentation.Rot90Augmentation import Rot90Augmentation
 from cryolo.utils import BoundBox, bbox_iou
 
 
@@ -35,20 +39,23 @@ class PatchwiseBatchGenerator(Sequence):
         images,
         config,
         patch_assigments,
+        cryolo_mode,
         shuffle=True,
         jitter=True,
         norm=None,
         overlap=100,
         name=None,
         train_times=1,
+        is_filament_training=False
     ):
-
+        self.is_filament_training=is_filament_training
         self.myname = name
         self.images = images
         self.imgdepth = 3
         self.config = config
         self.shuffle = shuffle
         self.jitter = jitter
+        self.cryolo_mode = cryolo_mode
         self.norm = norm
         self.patch_assigments = patch_assigments
         self.overlap = overlap
@@ -70,12 +77,18 @@ class PatchwiseBatchGenerator(Sequence):
                     self.patch_coords.extend([(i, p_x, p_y)] * train_times)
 
         random.shuffle(self.patch_coords)
+
         if len(img_first.shape) == 2:
             self.imgdepth = 1
-        elif np.all(img_first[:, :, 0] == img_first[:, :, 1]) and np.all(
-            img_first[:, :, 0] == img_first[:, :, 2]
-        ):
+        elif img_first.shape[0] == 3 and np.all(
+                img_first[:, :, 0] == img_first[:, :, 1]) and np.all(
+                img_first[:, :, 0] == img_first[:, :, 2]
+                ):
             self.imgdepth = 1
+        elif len(img_first.shape) == 3 and img_first.shape[0] > 3:
+            # Tomo data
+            self.imgdepth = 1
+
 
     def __len__(self):
         length = int(np.ceil(len(self.patch_coords) / self.batch_size))
@@ -102,7 +115,7 @@ class PatchwiseBatchGenerator(Sequence):
             dtype=np.float32,
         )  # input images
         b_batch = np.zeros(
-            (self.batch_size, 1, 1, 1, self.config["TRUE_BOX_BUFFER"], 4),
+            (self.batch_size, 1, 1, 1, self.config["TRUE_BOX_BUFFER"], 4 + self.is_filament_training),
             dtype=np.float32,
         )  # list of self.config['TRUE_self.config['BOX']_BUFFER'] GT boxes
         y_batch = np.zeros(
@@ -111,7 +124,7 @@ class PatchwiseBatchGenerator(Sequence):
                 self.config["GRID_H"],
                 self.config["GRID_W"],
                 self.config["BOX"],
-                4 + 1 + self.config["CLASS"],
+                4 + self.is_filament_training + 1 + self.config["CLASS"],
             ),
             dtype=np.float32,
         )  # desired network output
@@ -199,7 +212,11 @@ class PatchwiseBatchGenerator(Sequence):
                             float(self.config["IMAGE_H"]) / self.config["GRID_H"]
                         )  # unit: grid cell
 
-                        box = [center_x, center_y, center_w, center_h]
+                        if self.is_filament_training:
+                            angle = obj["angle"] / np.pi # normalize it between 0 and 1
+                            box = [center_x, center_y, center_w, center_h, angle]
+                        else:
+                            box = [center_x, center_y, center_w, center_h]
 
                         # find the anchor that best predicts this box
                         best_anchor = -1
@@ -217,11 +234,16 @@ class PatchwiseBatchGenerator(Sequence):
 
                         # assign ground truth x, y, w, h, confidence and class probs to y_batch
 
-                        y_batch[sample_index, grid_y, grid_x, best_anchor, 0:4] = box
+                        y_batch[sample_index, grid_y, grid_x, best_anchor, 0:4] = box[0:4]
                         y_batch[sample_index, grid_y, grid_x, best_anchor, 4] = 1.0
                         y_batch[
                             sample_index, grid_y, grid_x, best_anchor, 5 + obj_indx
                         ] = 1
+                        if self.is_filament_training:
+                            y_batch[
+                                sample_index, grid_y, grid_x, best_anchor, 5 + obj_indx + 1
+                            ] = angle # this line will lead to problems when multiple classes
+                            # would be predicted, as obj_indx then might be > 0
 
                         # assign the true box to b_batch
                         b_batch[sample_index, 0, 0, 0, true_box_index] = box
@@ -232,6 +254,7 @@ class PatchwiseBatchGenerator(Sequence):
             # assign input image to x_batch
             if self.norm is not None:
                 x_batch[sample_index] = self.norm(img)
+
         return [x_batch, b_batch], y_batch
 
     def on_epoch_end(self):
@@ -257,13 +280,18 @@ class PatchwiseBatchGenerator(Sequence):
         else:
             image_name = train_instance["filename"]
         patch_mode = region is not None
-        image = imagereader.image_read(image_name, region)
+        if "z" in train_instance:
+            # Get correct slice in case of tomograms
+            image = imagereader.image_read(image_name, region, use_mmap=True)
+            image = image[int(train_instance["z"]),:,:]
+        else:
+            image = imagereader.image_read(image_name, region, use_mmap=False)
         h = image.shape[0]
         w = image.shape[1]
 
         # if the filtered image size is not equals the input image size
         # that means we are in mixed-size mode. Set region randomly
-        if h != self.config["IMAGE_H"] or w != self.config["IMAGE_W"]:
+        if self.cryolo_mode == utils.CryoloMode.MIXED_ASPECT_RATIO:
             diffh = h - self.config["IMAGE_H"]
             diffw = w - self.config["IMAGE_W"]
             if diffh > 0.0:
@@ -295,7 +323,6 @@ class PatchwiseBatchGenerator(Sequence):
         all_objs = copy.deepcopy(train_instance["object"])
 
         if jitter:
-
             # scale the image
 
             scale = np.random.uniform() / 10.0 + 1.0
@@ -313,27 +340,28 @@ class PatchwiseBatchGenerator(Sequence):
 
 
             # flip the image
+
             flip_selection = np.random.randint(0, 4)
+            aug_flip = FlipAugmentation(flipping_mode=flip_selection)
+            image = aug_flip.transform_image(image)
 
 
-            flip_both = flip_selection == 1
-            flip_horizontal = flip_selection == 2
-            flip_vertical = flip_selection == 3
-
-
-            if flip_vertical:
-                image = np.flip(image, 1)
-            if flip_horizontal:
-                image = np.flip(image, 0)
-            if flip_both:
-                image = np.flip(np.flip(image, 0), 1)
-
             if image.shape[0] == image.shape[1]:
                 num_rots = np.random.randint(4)
             else:
                 num_rots = np.random.choice([2,4])
+            aug_rot = Rot90Augmentation(num_rotations=num_rots)
+            image = aug_rot.transform_image(image=image)
+
+
+
+            # rotation random degree maybe here?
+            custom_rotation = np.random.rand()>0.5
+            if custom_rotation:
+                custom_rotation_angle = np.random.rand() * np.pi/4
+                custrot = CustomRotationAugmentation(custom_rotation_angle)
+                image = custrot.transform_image(image)
 
-            image = np.rot90(image, k=num_rots, axes=(1, 0))
             is_grey = np.issubdtype(image.dtype, np.int8) or np.issubdtype(
                 image.dtype, np.uint8
             )
@@ -365,7 +393,6 @@ class PatchwiseBatchGenerator(Sequence):
                 offx = offx * w / float(self.config["IMAGE_W"])
                 offy = offy * h / float(self.config["IMAGE_H"])
 
-
         for obj in all_objs:
             if region is None:
                 is_in_region = True
@@ -395,54 +422,37 @@ class PatchwiseBatchGenerator(Sequence):
                 for attr in ["xmin", "xmax"]:
                     obj[attr] = obj[attr] - region_off_x
                     if jitter:
-                        obj[attr] = int(obj[attr] * scale - offx)
-                    obj[attr] = int(obj[attr] * float(self.config["IMAGE_W"]) / w)
+                        obj[attr] = obj[attr] * scale - offx
+                    obj[attr] = obj[attr] * float(self.config["IMAGE_W"]) / w
                     obj[attr] = max(min(obj[attr], self.config["IMAGE_W"]), 0)
 
                 for attr in ["ymin", "ymax"]:
                     obj[attr] = obj[attr] - region_off_y
                     if jitter:
-                        obj[attr] = int(obj[attr] * scale - offy)
-                    obj[attr] = int(obj[attr] * float(self.config["IMAGE_H"]) / h)
+                        obj[attr] = obj[attr] * scale - offy
+                    obj[attr] = obj[attr] * float(self.config["IMAGE_H"]) / h
                     obj[attr] = max(min(obj[attr], self.config["IMAGE_H"]), 0)
 
-                if jitter and (flip_vertical or flip_both):
-                    xmin = obj["xmin"]
-                    obj["xmin"] = self.config["IMAGE_W"] - obj["xmax"]
-                    obj["xmax"] = self.config["IMAGE_W"] - xmin
-
-                if jitter and (flip_horizontal or flip_both):
-                    ymin = obj["ymin"]
-                    obj["ymin"] = self.config["IMAGE_H"] - obj["ymax"]
-                    obj["ymax"] = self.config["IMAGE_H"] - ymin
+                if jitter and aug_flip:
+                    obj = aug_flip.transform_coords([obj], [self.config["IMAGE_W"],self.config["IMAGE_H"]])[0]
 
                 if jitter and num_rots > 0:
-                    hcenter = float(self.config["IMAGE_H"]) / 2
-                    wcenter = float(self.config["IMAGE_W"]) / 2
+                    obj = aug_rot.transform_coords([obj], [self.config["IMAGE_W"],self.config["IMAGE_H"]])[0]
 
-                    for i in range(num_rots):
-                        obj["xmin"] = obj["xmin"] - wcenter
-                        obj["xmax"] = obj["xmax"] - wcenter
-                        obj["ymin"] = obj["ymin"] - hcenter
-                        obj["ymax"] = obj["ymax"] - hcenter
+                if jitter and custom_rotation:
+                    obj = custrot.transform_coords([obj], image_dims=[self.config["IMAGE_W"], self.config["IMAGE_H"]])[0]
+                    centerx = obj["xmin"] + (obj["xmax"] - obj["xmin"]) / 2
+                    centery = obj["ymin"] + (obj["ymax"] - obj["ymin"]) / 2
 
-                        help = obj["xmin"]
-                        obj["xmin"] = -1 * obj["ymin"]
-                        obj["ymin"] = help
+                    if centerx < 0 or centerx>self.config["IMAGE_W"] or centery < 0 or centery>self.config["IMAGE_H"]:
+                        continue
 
-                        help = obj["xmax"]
-                        obj["xmax"] = -1 * obj["ymax"]
-                        obj["ymax"] = help
-
-                        obj["xmin"] = obj["xmin"] + wcenter
-                        obj["xmax"] = obj["xmax"] + wcenter
-                        obj["ymin"] = obj["ymin"] + hcenter
-                        obj["ymax"] = obj["ymax"] + hcenter
-
-                        # Swap xmin and xmax
-                        help = obj["xmax"]
-                        obj["xmax"] = obj["xmin"]
-                        obj["xmin"] = help
                 obj_is_region.append(obj)
+        #print(image.dtype)
+        if False:
 
+            print(image_name, int(train_instance["z"]), len(all_objs), len(obj_is_region))
+            if len(obj_is_region)>0:
+                for kk in range(len(obj_is_region)):
+                    print(obj_is_region[kk])
         return image, obj_is_region
